﻿/*
   Copyright 2012, Andrew Polar

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Reflection;
using wordprocessing;

namespace dicluster
{
    class EntryPoint
    {
        static int getMode(int[] x) 
        {
            int max = x[0];
            foreach (int y in x)
            {
                if (y > max) max = y;
            }
            int size = max + 1;
            int[] z = new int[size];
            for (int k = 0; k < z.Length; ++k)
            {
                z[k] = 0;
            }
            foreach (int y in x)
            {
                ++z[y];
            }
            int pos = 0;
            max = z[pos];
            int cnt = 0;
            foreach (int y in z)
            {
                if (y > max)
                {
                    max = y;
                    pos = cnt;
                }
                ++cnt;
            }
            return pos;
        }

        static void Main(string[] args)
        {
            string rootFolder = @"..//..//..//..//PATENTCORPUS128";  //This location of data to be processed, make sure it points to your data.
            string extension = "*.txt";
            string dataFolder = "data";  //name of the folder (local to executable) where processed data is stored
            //Next are two names of files with processed data. 
            //ProcessedFilesList.txt is ASCII file with list of found and processed files. 
            string filesList = "ProcessedFilesList.txt";  
            string dataFile = "DocWordMatrix.dat";
            string path = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
            dataFolder = Path.Combine(path, dataFolder);
            filesList = Path.Combine(dataFolder, filesList);
            dataFile = Path.Combine(dataFolder, dataFile);

            DirectoryInfo di = new DirectoryInfo(rootFolder);
            if (!di.Exists)
            {
                Console.WriteLine("The data folder not found, please correct the path");
                return;
            }

            DateTime start = DateTime.Now;
            //
            //This is training sample, it is good only PATENTCORPUS128.
            //For other data training sample has to be changed.
            //The training sample is recored in ProcessedFilesList.txt file in data directory.
            //Example of two lines from ProcessedFilesList.txt:
            //3, 0, C:\APOLAR\UTILITIES\PATENTCORPUS128\C399P7058343.txt
            //4, -1, C:\APOLAR\UTILITIES\PATENTCORPUS128\C399P7164866.txt
            //first number is index of processed file, second number is known category if >= 0 or
            //indicator that file to be categorized if < 0;
            //Make sure the training set is correct for the particular order of processed files.
            //It can be found in the file ProcessedFilesList.txt.
            //
            int N = 128;
            int[] nChart = new int[N];
            for (int i = 0; i < N; ++i)
            {
                nChart[i] = -1;
            }

            //This is training sample. Line nChart[16] = 1; nChart[21] = 1; nChart[30] = 1; means
            //that files 16, 21 and 30 in the list belong to same category that conventionally marked 1. 
            nChart[0] = 0; nChart[4] = 0;  nChart[8] = 0; nChart[14] = 0; nChart[15] = 0;
            nChart[16] = 1; nChart[21] = 1; nChart[30] = 1;
            nChart[32] = 2; nChart[34] = 2; nChart[35] = 2;  nChart[46] = 2;
            nChart[48] = 3; nChart[51] = 3; nChart[56] = 3;  nChart[60] = 3;
            nChart[64] = 4; nChart[70] = 4; nChart[71] = 4;  nChart[76] = 4;
            nChart[80] = 5; nChart[82] = 5;  nChart[90] = 5;
            nChart[100] = 6; nChart[105] = 6;  nChart[108] = 6;
            nChart[118] = 7; nChart[122] = 7; nChart[123] = 7;  nChart[124] = 7;

            //The following data are needed for automatic estimation of the accuracy
            //This data is good only for PATENTCORPUS128.
            int nCategories = 8;
            //This is expected result. 
            //The expected result is category sitting in file index array.
            int[] nExpected = new int[128];
            int nCatSize = 16;
            int nWhich = 0;
            for (int i = 0; i < 128; ++i)
            {
                nExpected[i] = nWhich;
                if ((i + 1) % nCatSize == 0)
                {
                    ++nWhich;
                }
            }

            //Here we start word processing and building document-word matrix
            bool useStemming = true;  //This parameter enables or disables stemming.
            WordProcessing wp = new WordProcessing(dataFolder, filesList, dataFile);
            wp.processFiles(rootFolder, extension, nChart, useStemming);
            //At this point document-word matrix is ready and we test clustering technologies.

            int[,] matrix = new int[5, 128];
            int[] res = new int[128];

            //1. The following is test for Naive Bayes
            NB nb = new NB();
            if (!nb.processNB(filesList, dataFile, nCategories))
            {
                Console.WriteLine("Failed to process Naive Bayes");
            }
            else
            {
                nb.showFilesInClusters();
                Console.WriteLine("Accuracy for NB {0}", nb.computeF_measure(nExpected, nCategories));
                Console.WriteLine();
                res = nb.getResult();
                for (int i = 0; i < res.Length; ++i)
                {
                    matrix[0, i] = res[i];
                }
            }
            //End of Naive Bayes

            //2. This is test for HAC
            HAC hac = new HAC();
            hac.populateFileAndTabooLists(filesList);
            hac.executeHAC(dataFile, nCategories);
            Console.WriteLine("Accuracy for HAC {0}", hac.computeF_measure(nExpected, nCategories));
            Console.WriteLine();
            res = hac.getResult();
            for (int i = 0; i < res.Length; ++i)
            {
                matrix[1, i] = res[i];
            }
            //End of HAC

            //3. Next is test for orthogonal decomposition invented by Andrew Polar. 
            Orthogon ortho = new Orthogon();
            ortho.BuildOrthogonalClassifier(dataFile, filesList);
            ortho.ClusterDocuments(dataFile);
            Console.WriteLine("Accuracy for Polar orthogonalization {0}", ortho.computeF_measure(nExpected, nCategories));
            Console.WriteLine();
            res = ortho.getResult();
            for (int i = 0; i < res.Length; ++i)
            {
                matrix[2, i] = res[i];
            }
            //End of orthogonal decomposition 

            //4. Next is k-means mixed with orthogonalization decomposition invented by Andrew Polar.
            KmeansPlusPO kmean = new KmeansPlusPO();
            kmean.BuildOrthogonalClassifier(dataFile, filesList);
            kmean.ClusterDocuments(dataFile);
            kmean.assignPlusUpdate(dataFile);
            kmean.assignPlusUpdate(dataFile);  //two runs for k-means is enough
            Console.WriteLine("Accuracy for k-means + PO {0}", kmean.computeF_measure(nExpected, nCategories));
            Console.WriteLine();
            res = kmean.getResult();
            for (int i = 0; i < res.Length; ++i)
            {
                matrix[3, i] = res[i];
            }
            //End of k-means with orthogonalization 

            //5. Next is Random Forest with orthogonalization decomposition invented by Andrew Polar. 
            RFPlusPO rf = new RFPlusPO();
            if (!rf.buildTrainingVectors(dataFile, filesList))
            {
                Console.WriteLine("Failed to build training matrix");
            }
            else
            {
                rf.buidForest(50);
                rf.processVectors(dataFile, nCategories);
                Console.WriteLine("Accuracy for RF + PO {0}", rf.computeF_measure(nExpected, nCategories));
                res = rf.getResult();
                for (int i = 0; i < res.Length; ++i)
                {
                    matrix[4, i] = res[i];
                }
            }
            //End of RF

            DateTime end = DateTime.Now;
            TimeSpan duration = end - start;
            double time = duration.Minutes * 60.0 + duration.Seconds + duration.Milliseconds / 1000.0;
            Console.WriteLine("Total processing time {0:########.00} seconds\n", time);

            //This is utilization of several different technologies, getting best result from simple voting
            Console.WriteLine("Simple voting");
            int[] x = new int[5];
            for (int i = 0; i < 128; ++i)
            {
                int cnt = 0;
                Console.Write(" {0:D} - ", i);
                for (int j=0; j<5; ++j) 
                {
                    Console.Write(" {0:D} ", matrix[j, i]);
                    x[cnt++] = matrix[j, i];                                         
                }
                res[i] = getMode(x);
                Console.Write(" --expected-- {0:D}", nExpected[i]);
                Console.WriteLine();
            }
            Console.WriteLine();
            
            Console.WriteLine("Accuracy for simple voting {0}\n", Helper.computeF_measure(nExpected, res, nCategories, "simple voting"));
        }
    }
}
